; Sort $descs when Zone stabilizes
; RCK when commands work.
comment~ TraceBack starts knowing execution is in file Z.File, at the
level of Z.Level.  CursorSi will say where the execution is in the relevant
Pcode. Routine Cursor2Clause will work out what clause & line number that
corresponds to. (Since the clause is shown in the traceback it looks like
we don't need to also have the name of the routine. However, there are
messages with <name> inserts so keep the name for that purpose.)

The callers are found on the UpLevel chain through the level blocks. There
is an IsExternal flag on the top level for a file, indicating that the caller's
L.File should be copied to Z.File when the caller becomes current.

~
; Copyright Formcroft Ltd 1997-1999
 .model small.486.code
 include always.inc; Assembler shorthands like 'Bptr' for 'byte ptr'.
 include declares.inc; Symbolic constants, macros, records, structures.
 PUBLIC MakeMsg
 PUBLIC ShowMsg
 PUBLIC TargetMsg; Used by ERRORTEXT
 EXTERN BifNames:near; Label for lookup of bif insert.
 EXTERN Keys:near; Label to help compaction - use words from keyword table.
 EXTERN Exit:near; Label to exit.
; See the MSGC utility amongst my WARP programs for making of cmp.inc
; (That takes IS.MSG as uncompressed input. Descriptions order to be matched.)
; Avoid first offset in compressed msgs being zero. Not edited.
                                                               include cmp.inc
comment~ There is either going to be a message (and termination) or SIGNAL ON
SYNTAX.  The description inserts in some messages are not specified by the
Standard.  See our raw message text for the prose of these.  The numbers there
must match $Desc... values.

Fairly complicated, with layers of routines, because of decompression and
inserts.
~

;------------------------------------------------------------------------------
; MakeMsg - Construct the message specified by ax.
; With adjustments for special cases.
;------------------------------------------------------------------------------

MakeMsg:push bxmv ds,cs
; There are actually two or three messages as far as the ErrorText is
; concerned. A 0.1 or 0.2 to locate the problem, then a Z.Insert.MajMin,
; then maybe an implementation provided Description.
; (A 0.3 is done by calling TargetMsg directly)
; The component messages are made in sequence so that values for inserts
; can change between components.
; Assert - nothing Insert does will need Z.Stack increase.
; We cannot use Z.Serial as workspace because Z.Insert.TokPos points into it.
 mov StackDi,Z.StackUp
 mov Z.MsgSoFar,StackDimov Z.MsgBegin,StackDi
 add Z.Stack,$MaxPreInsert+5*$Limit_Insertcall StackSafe
 mov ax,MkMajMin(0,1)
; !! inc ax when interactive
 call TargetMsg
 move Z.InsLine,Z.ClauseLinemov ax,Z.Insert.MajMincall TargetMsg
; Luckily, the <description> is always last in the messages. So it just sets
; a flag, acted on here.
 btr Z.Cancel,Descriptionjnc @F
 mov ah,$Descriptionsmov al,Z.Insert.Desccall TargetMsg
@@:pop bxret

;------------------------------------------------------------------------------
; TargetMsg - Expand the message specified by ax.
; Flag says whether inserts are to be expanded.
;------------------------------------------------------------------------------
TargetMsg:mov Z.TargetMajMin,ax
CharCx is <cx>; Testing insert letter
CharCl is <cl>; Testing insert letter
CharAl is <al>; Testing insert letter
CurrentByteAx is <ax>; AH is kept zero. (LODS used)
LowCurrentByteAl is <al>
 mv es,ss
; Now the scan. Start at -1 because there is a message 0.
; Put the msg without inserts somewhere it can be copied left while inserting.
 mov MajMinDx,MkMajMin(-1,0)mov di,Z.MsgSoFaradd di,$MsgRoompush di
 z CurrentByteAxz DepthBl
 mov si,offset MsgcPackedmov cx,MsgcUnitscall MovePart
; Raw message is now at [Z.MsgSoFar]+$MsgRoom..., ending before di.
 mov Z.PreZi,dipop si; [Z.MsgSoFar]+$MsgRoom
; Uppercase the first letter.
 cmp Bptr[si+bp],'<'je @Fand Bptr[si+bp],0DFh@@:
 mov di,Z.MsgSoFar; Next copy is to Z.MsgSoFar
 mv ds,ss; Shuffle ds to save prefixingjmp WhileInsert
LoopInsert:lodsbcmp CharAl,'<'jne NotIns
; For the builtin, inserts are left unaltered.
 Qry Z.Cancel,ForBIFjnz NotIns
; The <> are not kept
 lodsb
; Insert does switches on what sort of insert.
 z cxmov cl,alsub CharCl,'a'-1;
 push sicall Insertmv ds,sspop si
 ; Skip past next '>'
@@:lodsbcmp CharAl,'>'jne @Bjmp WhileInsert
NotIns:stosb; Plain copy when not inserting.
WhileInsert:cmp si,Z.PreZijb LoopInsert
 mv ds,csmov Z.MsgSoFar,diret; From TargetMsg

Insert:
; We are set up with es=ss so that movsb will be into message buffer.
; We are set up with ds=ss which can be changed.
; ax & & bx & dx free to use. cx free after finding which insert.
; si to first letter of insert type, free for reuse.
 loop Nota;<argnumber>
 movzx ax,Z.Insert.ArgNumjmp ShowAx
Nota:loop Notb;<bif>
; Reverse lookup - Bif number to spelling.
 mv ds,csmov si,offset BifNames; Where spellings are laid out.
PerBif:mov cx,si
@@:lodsbsub al,$BifTagjb @B
 cmp al,Z.InsBifjne PerBif
 xchg cx,sisub cx,sidec cx
MoveQuit:rep movsbret
Notb:loop Notc;<char>mov al,Z.Insert.Charstosbret
Notc:loop Notd;<description>On Z.Cancel,Descriptionret
Notd:dec CharCxdec CharCxdec CharCx
 loop Noth;<hex-encoding>movzx ax,Z.Insert.Hexmov Z.Radix,16jmp ShowAx
; Code moved to bring loop Notk into range.
; It takes about 143 bytes of code to do this keywords insert.  Actually
; expanding in the original message might do better - there are 64 elements.
DoneArray:
 TestReg axjnz AnotherPassmv es,ssmov di,Z.Insert.KeysListmov bx,sp
InsKey:
 mov si,ss:[bx]mov cl,[si]and cl,mask KeyLeninc sirep movsb
 inc bxinc bxmov cx,dxsub cx,bxjz InsDone
; Early separators are ", "
 mov ax," ,"stoswcmp cl,2jne InsKey
; Last separator is ", or " or " or " depending if total list was two items.
 dec dimov cx,bxsub cx,spcmp cl,2jne @Fdec di@@:
 mov ax,"o "stoswmov ax," r"stoswjmp InsKey
InsDone:mov sp,dx; recover data space used.ret
Noth:dec CharCxdec CharCx
 loop Notk;<keywords>
; Everything with this insert will be a error state preceeded by keys offset.
 mv es,csmv ds,cs
 mov si,Z.Insert.KeysListmov si,[si-2]; To keyword list
 mov Z.Insert.KeysList,di; Just a place to save it.
 z axmov dx,sp
; Run down the list in keyword table, pushing to make an array of offsets.
; First byte is data. Second byte is data including keyword length.
@@:add si,axinc sipush si
 lodsband al,mask EndList+mask KeyLen; FORM has KeyFlags
 btr ax,EndListjnc @B
AnotherPass:; Stack from sp up thru dx has the pointers to keywords.
 z axmov bx,sp
AnotherPair:; Compare this one with one above.
 mov cx,dxsub cx,bxdec cxdec cxjz DoneArray; One left not enough for pair.
 mov si,ss:[bx]inc bxinc bxmov di,ss:[bx]
; Compare at shorter length
 mov cl,[si]and cl,mask KeyLencmp cl,[di]jb @Fmov cl,[di]@@:
 inc siinc direpe cmpsbjbe AnotherPair
; Exchange needed.
 mov di,ss:[bx]xchg di,ss:[bx-2]mov ss:[bx],diinc axjmp AnotherPair
Notk:loop Notl;<linenumber>mov ax,Z.InsLine
ShowAx:call Ax2DsSiCxjmp MoveQuit
Notl:dec CharCx
 loop Notn;<name>  NYI SIGNAL VALUE dud
 mov ds,Z.Symbolsmov si,Z.Insert.Namedmovzx cx,Bptr[si]inc sijmp MoveQuit
Notn:loop Noto;<optionslist><operator>  <operation>!!
 cmp Bptr[si+bp+8],'>'je NotnTemp
 cmp Bptr[si+bp+7],'>'jne @F
NotnTemp:
; A synchronisation has been done when messages are put out, so we can use
; Z.InsOperator.
 mov si,Zshape.InsOperatormovzx cx,Bptr[si+bp]inc sijmp MoveQuit
@@:mov si,Z.Insert.Options
@@:lodsb cs:Bptr[si]stosbcmp si,Z.Insert.OptsHijbe @B
SomeRet:ret
Noto:loop Notp;<position>mov ax,Z.Insert.Positionjmp ShowAx

Notp:dec CharCxdec CharCx
 loop Nots;<source><sourceline()>
 cmp Bptr[si+bp+5],'>'jne Notsr
 lds bx,Z.InsSourcelea si,Jbx.Keymov cx,Jbx.KeyLenjmp MoveQuit

Notsr:lds bx,Z.SysFilemov ax,Wptr F.Linesjmp ShowAx

Nots:loop Nott;<token>
 Qry Z.Cancel,TokIsValuejnz AsValue
 mov si,Z.Insert.TokPosmov cx,Z.Insert.TokLen
LimitedInsert:
 z bx; "..." indicator. Should depend on quotes but that is pedantic.
 cmp cx,$Limit_Insertjna @Fmov cx,$Limit_Insertinc bx@@:
 JcxzQ LiA; Failsafe uninitialized
NotsLoop:lodsb
 ; If keyword, uppercase. (Where does it say that?)
 cmp Z.Insert.Major,27jne @Fand al,0DFh@@:
 stosbloop NotsLoop
LiA:dec bxjnz @Fmov al,'.'stosbstosbstosb@@:ret
Nott:; <values> are usually Rexx variable values but not always.
 cmp Z.TargetMajor,0jnz @F
; In this case, <value> means Z.Insert.MajMin, the number of the error
; Have to make Z.Insert.MajMin in a readable form. Zero minor codes are not shown.
 movzx ax,Z.Insert.Majorcmp Z.Insert.Minor,0jz ShowAx
 call Ax2DsSiCxrep movsbmov al,'.'stosbmovzx ax,Bptr Z.Insert.Minor
JmpShowAx:jmp ShowAx
; In the rest of the cases a value is on the soft stack.
; There are nine cases where messages have two <value> inserts. Mostly they
; are adjacent values on the stack.
; However 26.8 and 33.1 need DIGITS() as a first <value>. It would be nice
; to trick that by making the message have, say, a <position> insert instead
; but that would upset bif ERRORTEXT.
; So "DIGITS() first" is indicated by non-zero Z.Insert.Digits.
; 26.11 & 26.12 need DIGITS as third <value>, which caller puts on the stack.
AsValue:
@@:z axxchg ax,Z.Insert.DigitsTestReg axjnz JmpShowAx
 push dimov di,Z.Insert.Value
 add Z.Insert.Value,$StrideVcall ToDsSiCxpop dimv es,ssjmp LimitedInsert
;Insert ends

; MovePart - Expanding a piece of the compressed messages.
; All the text for all messages is scanned but it is not expanded until the
; relevant message is reached.
MovePart:; Recursive; Assert PartLenCx not zero.
 cmp bl,50jna MPloopError49Bp;Unexpected recursion in message production.
MPloop:lodsb
; Lowest values are three special cases.
 cmp LowCurrentByteAl,0jne NotKeyword
; convert from keyword specified in next byte to keyword itself.
; MASM bug pad ......................................................
 lodsbdec PartLenCxpush PartLenCxpush ScannerSicmp TheMsgBh,0jz UnitDone
 mov ScannerSi,offset Keysmov KeywordOffsetCx,CurrentByteAxadd cx,cxadd cx,si
; Loop until ScannerSi is at the one we want.
 dec ScannerSi
@@:inc ScannerSiinc ScannerSimov al,Bptr[ScannerSi]and al,15; length
 cmp ScannerSi,KeywordOffsetCxjae @Fadd ScannerSi,CurrentByteAxjmp @B
; Literal of keyword follows.
@@:z PartLenCxmov LowPartLenCl,alinc ScannerSirep movsbjmp UnitDone
NotKeyword:push PartLenCxpush ScannerSi
 dec LowCurrentByteAljz BumpMinorDldec LowCurrentByteAljz BumpMajorDh
 cmp LowCurrentByteAl,MsgcFragsLo-2jb Unique
; A fragment to be expanded, details in an array.
 mov si,CurrentByteAx; Index to Ushort
 add si,simov PartLenCx,[si]+(offset MsgcIndex)-(MsgcFragsLo-2)*2
; This UShort has position and length packed in it 12:4.
 mov ScannerSi,PartLenCxshr ScannerSi,4add ScannerSi,offset MsgcPacked
 and PartLenCx,0Fhinc2 PartLenCx
 inc DepthBxcall MovePartdec DepthBx; Recursion
UnitDone:pop ScannerSipop PartLenCxLoopQ MPloopret

Unique:; Reached the depths - a character from amongst the uniques in original.
 cmp TheMsgBh,0jz UnitDone
 mov si,CurrentByteAxadd si,offset MsgcUniques-1movsbjmp UnitDone

BumpMajorDh:inc MajorDhz MinorDljmp @F
BumpMinorDl:inc MinorDx
@@:; TheMsgBh reflects whether we are in the message we want.
 z TheMsgBhcmp MajMinDx,Z.TargetMajMinjne @Finc TheMsgBhjmp UnitDone
@@:jb UnitDone
; Can't take advantage of passing over the target until recursion unwound.
 TestReg DepthBljne UnitDone
 pop ScannerSipop PartLenCx; Alternate pop to that at UnitDone
 ret

ShowMsg:; Show it, in chunks up to $MsgChunk
; A touch of over-engineering because ANSI says don't worry about whether
; message will get words split across lines.
ResidueAx is <ax>
ChunkCx is <cx>
SoFarDx is <dx>
 mv ds,ssmov bx,Z.Stderr; On this file
 mov SoFarDx,Z.MsgBegin; From here to di
ShowMsgLoop:mov ax,disub ResidueAx,SoFarDx; Length remaining
 mov ChunkCx,$MsgChunkcmp ResidueAx,ChunkCxjna ShowMsgL
; Overlong. Break at rightmost blank in chunk.
 mov ax,ChunkCxmov si,SoFarDxadd si,ChunkCx
@@:cmp Bptr[si],' 'je ShowMsgLinedec siloop @B
ShowMsgL:mov ChunkCx,ax
ShowMsgLine:mov ah,$DosWriteint 21h; Takes DS:DX & CX
 add SoFarDx,ChunkCxcmp SoFarDx,dijne @Fret@@:
; Prefix the rest with newline.
 dec SoFarDxdec SoFarDxmov si,SoFarDxmov Wptr[si],0A0Dh
 jmp ShowMsgLoop
 end
