;------------------------------------------------------------------------------
; Processing for tracing proceeds clause by clause.
;------------------------------------------------------------------------------
comment~ Rexx tracing shows the clause before executing but doesn't pause until
a clause ends.  Showing and executing are different activities - all the clause
will be shown even if the raising of a condition means that it isn't all
executed.

To do the execution aspects of tracing, extra Pcode is made transiently on a
clause-by-clause basis.  This extra Pcode has the same format and opcodes as
ordinary (fast-form, original) Pcode but with an extra opcode $pTraceEvent used.

The sequence of activity is:

a) Fast-form Pcode is made for the whole source file and execution begins.

b) After a TRACE instruction (or TRACE bif) there is a
test for whether tracing is needed.  When the need for tracing is detected a
non-producing parse is done, (Synch), to put the source position and the
execution position in step.

c) Pcode, including trace events, is generated for the clause. Any jump targets
in this transient Pcode take their values from the original Pcode. During this
generation the limit of the source relating to this clause is noted.
See ClauseSourceEnd.
Also the offset of the original Pcode following the clause is noted.

d) The source is shown if the trace setting requires that.

e) The transient Pcode is executed. The effect is the same as ordinary Pcode
plus writing to the trace file when $pTraceEvent opcodes are executed.
However, the operations are monitored for whether they produce jumps.

e) The trace event for end-of-clause will be executed.

There are some complications:

When the TRACE bif starts the trace, there will be no source shown for that
clause, so TraceSource is irrelevant.  It is marked negative to indicate this.

When the TRACE instruction starts the trace, we do not want the Action91 at
the end of that clause to compile an end-of-clause event.  The setting of
DualGen has to be delayed until after, by the DualPending flag.
~

comment~ When tracing there will be transient separate Pcode made for each
clause and it will be executed by XCycle.  It uses the same set of
Pcode operators as are available to XCycle but also has those which come here.
Code here checks after each operator
is executed to see if there have been a jump.
(NYI.The idea is that fallthru could avoid a synch but this version synchs
anyway.)
~

Monitor:; The same action as for XCycle operations. See label Operator.
 movzx bx,al
 mov Z.Stack,StackDi; Record of how much stack in use needed for Compaction.
 call cs:(PcodesBase$-1)[bx]
; The CursorSi now will tell us whether the operation resulted in a jump.
; A jump target is in the original Pcode so will be below the transient Pcode.
 cmp CursorSi,Z.DualCodejnb XCycle
; We jumped out of that clause. That ends the clause and tells us where to
; start the next.
JumpedOut:
 mov Z.Cursor,CursorSi
; Clause can end by reason of jump or by $TraceEndClause event at end.
ClauseRunEnd:
 Off Z.Cancel,ParseTrace
 mov CursorSi,Z.Cursor; Next clause, on the original.
; If it is the $pExit synthesised on the end of the original program then there
; is no corresponding source.
 lea ax,1[CursorSi]
 cmp ax,Hg.Original
 je StartExecutionY; Run the original $pExit
DualSwitch:; CursorSi is ready for the next clause.
 mov sp,Z.XCycleResetSp
; There may be no reason to stay in (or enter) clause-by-clause mode.
 call HeritageDimov al,L.TraceLetterpop StackDi
; Make sure of ShowClauseId each time tracing starts up.
 Qry Z.Gen,Dualjnz @Fz Z.TraceLine@@:
 cmp al,'N'je XCyclecmp al,'O'je XCycle
; Do the next clause by dual method.
 mov Z.Synch,CursorSi
 mov Z.Gen,mask Dual
 mov Z.Cursor,$StrideH
 call Compiling
; Bounds of the source will have been a compiling by-product.
; Convert the current letter to a set of flags.
 call HeritageDi
 movzx bx,L.TraceLettermov dl,cs:$TraceLetter[bx-'A']pop StackDi
 test dl,$TraceAjz NotTraceA
; ELSE has special treatment.
 Qry Z.Program,ShowELSEjnz MonNotThen
; We want the lines of source between Z.TraceSource and Z.TraceEnds.
; THEN is special. We had to push on over it in the source but it isn't
; supposed to show if the THEN clause isn't executed.
 mov CursorSi,Z.DualCode
 inc CursorSi; Over $pTraceEvent
 cmp Bptr gs:[CursorSi],$pThenje @F
 cmp Bptr gs:[CursorSi],$pWhenjne MonNotThen
@@:inc CursorSi; Over $pThen
; Have to run THEN before (perhaps) showing it.
 call Then$$
 On Z.Program,ShowELSE
 cmp CursorSi,Z.DualCodejb JumpedOut; Nothing shown
 mov Z.DualCode,CursorSi
 Off Z.Program,ShowELSE
MonNotThen:
; ELSE is special. The unconditional jump it puts at the end of a THEN leg,
; if executed, does not have any corresponding source shown.
 cmp Bptr gs:[CursorSi],$pJumpje ExecuteDual; Showing no source.
; Our window on the source is Z.Relative onwards for $SourceBufferSize-2.
; No need to refetch if what we want is in the window.
 mov ecx,Z.TraceSource
 cmp Z.SerialNow,$SerialSourcejne @F
 mov edx,Z.Relativecmp ecx,edxjb @F
 add edx,$SourceBufferSize-2cmp edx,Z.TraceEndsja NoNeedSource
@@:mov Z.Relative,ecxmov al,$SerialSourcecall Serialize
NoNeedSource:mov esi,ecxsub esi,Z.Relativeadd si,Zshape.Serial
; Bytes to show start at SS:SI
 mov Dptr(Z.Tag+1),StringOrder('*','-','*',' '); For first ShowClauseId
; Dont show leading $Lf
 mv ds,sscmp Bptr[si],$Lfjne @Finc siinc ecxz Z.TraceSourceColumn@@:
 mov dx,sicmp ecx,Z.TraceEndsjnb NotTraceA
TSloop:lodsbcmp al,$Lfjne TSnext
 call ShowLine; dx up to si
TSnext:inc ecxcmp ecx,Z.TraceEndsjne TSloop
; If there is anything to show at this point it will not end with a line-feed.
; One has to be added.
 cmp si,dxje NotTraceA
 call ShowLine
 mv ds,csmov dx,Yshape.CrLfmov cx,2
 mov ah,$DosWriteint 21h
NotTraceA:
  btr Z.Program,ShowELSEjnc ExecuteDual
; Go again with same Z.Synch but flag off.
 mov CursorSi,Z.Synchjmp DualSwitch

; Code with trace events starts at Z.DualCode, ends with $TraceEndClause.
ExecuteDual:mov CursorSi,Z.DualCodejmp StartExecutionY

; Convert letter to flags, allowing for 'C' includes 'E' etc.
$TraceO equ 0
$TraceB equ 1; TraceBack
$TraceA equ 2
$TraceC equ 4
$TraceE equ 8
$TraceF equ 16
$TraceI equ 32
$TraceL equ 64
$TraceR equ 128
if Solid
; Whatever the letter, traceback is enabled.
$TraceLetter db $TraceA+1db ?db $TraceC+$TraceE+1db ?
 db $TraceE+1db $TraceF+1db ?db ?db $TraceA+$TraceI+1db ?db ?
 db $TraceL+1db ?db $TraceF+1db $TraceO+1db ?db ?db $TraceA+$TraceR+1
endif

; Convert trace event number to what letter (mask) enables it.
if Solid
$TrEv equ $
 org $TrEv+3*$TraceClause/2
 db $TraceA+$TraceI+$TraceR
 db "*-" ; Source, as with All
 org $TrEv+3*$TraceExpr/2
 db $TraceR
 db ">>" ; Results
 org $TrEv+3*$TraceBY/2
 db $TraceR
 db ">+" ; BY
 org $TrEv+3*$TraceDot/2
 db $TraceR
 db ">." ; Parse .
 org $TrEv+3*$TraceAssign/2
 db $TraceR
 db ">=" ; Assign, parse assign.
 org $TrEv+3*$TraceCompound/2
 db $TraceI
 db ">C" ; Parts of a compound reference.
 org $TrEv+3*$TraceLit/2
 db $TraceI
 db ">L" ; Literal
 org $TrEv+3*$TracePrefix/2
 db $TraceI
 db ">P" ; Prefix operator
 org $TrEv+3*$TraceOperator/2
 db $TraceI
 db ">O" ; Other operator
 org $TrEv+3*$TraceFunction/2
 db $TraceI
 db ">F" ; Function reference.
 org $TrEv+3*$TraceVar/2
 db $TraceI
 db ">V" ; Variable
 org $TrEv+3*$TraceError/2
 db $TraceE
 db "++" ; Return code
 org $TrEv+3*$TraceFailure/2
 db $TraceF
 db "++" ; Return code
 org $TrEv+3*$TraceBack/2
 db $TraceB; Highest
 db "*T" ; Traceback
; Also "*-"  and "*," for source.
endif

TraceEvent$$:lodsb gs:[CursorSi]
; Odds are Pcodes, evens are particular causes of tracing.
 test al,1jnz Monitor
; $TraceEndClause is special because it ends the transient code.
 cmp al,$TraceEndClauseje ClauseRunEnd
; Just set a flag when PARSE.
 cmp al,$TraceParsejne @FOn Z.Cancel,ParseTraceret@@:

; Normal entry is from XCycle but there is this one from mid-parse. (With more
; registers in use. DsSiCx )
TraceEventX:push CursorSi; (SI is not CursorSi when from mid-parse)
; Convert the current letter to a set of flags.
 call HeritageDi
 movzx bx,L.TraceLettermov dl,cs:$TraceLetter[bx-'A']pop StackDi
; Note whether letter (mask) enables this event.
; Event of loading Var becomes event of loading Lit if variable is dropped.
 btr Z.Cancel,WasDroppedjnc @Fmov al,$TraceLit@@:
 cbwmov bx,axshr bx,1add bx,axand dl,Bptr cs:$TrEv[bx]jz TrEvRet
; Note the tag for the line to be shown.
 mov bx,cs:$TrEv[bx+1]mov Wptr Z.Tag+1,bxmov Z.Tag+3,bl
; DL now says what is to be shown.
 cmp al,$TraceAssignjne NotAssign$
; The assign in parse does not have a value of the stack. It is DS:SI:CX
 Qry Z.Cancel,ParseTracejz ShowTop
@@:; We need some reasonable place on the stack, to make it similar to others.
 mov StackDi,Z.Stackadd StackDi,2*$StrideVcall VasDsSiCxjmp ShowTop
NotAssign$:cmp al,$TraceDotje @B
 test dl,$TraceC or $TraceI or $TraceRjz NotTop
ShowTop:
; The value to be shown is on the stack.
 mov bx,Z.Stdoutcall ShowClauseId
 mv ds,csmov dx,Yshape.DqCrLfmov cx,1
 mov ah,$DosWriteint 21h; Takes ds:dx:cx to file bx
; We want to leave V. the way we found it.
; But cannot but V. back to numeric until showing done.
 test V.Dtype,$Testpushfcall ToDsSiCxpush cx
 cmp cx,$Limit_Insertjna @Fmov cx,$Limit_Insert@@:
 mov dx,si
; Better not let things like e-o-f show.
; How to do that without altering version in memory?
; Maybe show char by char if troubled? NYI
 mov ah,$DosWriteint 21h; Takes ds:dx:cx to file bx
 mv ds,csmov dx,Yshape.DqCrLf
; "..." preceeds CrLf, if truncation.
 pop cxcmp cx,$Limit_Insertmov cx,2jna @Fsub dx,3mov cx,5@@:
 mov ah,$DosWriteint 21h
 popfjnz @Fcall NumDxBx@@:
 RetSi

NotTop:

 test dl,$TraceE or $TraceFjz NotRC
; Show RC=
 mv ds,csmov dx,offset RCismov cx,3mov ah,$DosWriteint 21h
; Move to stack
 UpMove V,Vf[$?RCx]jmp ShowTop

NotRC:Error49Bp; L & T

TrEvRet:RetSi

RCis db "RC="

ShowClauseId:; BX is target file.
 push sipush dxpush cx
; TraceWidth to dl - how many columns for the number.
 push bxlds bx,Z.SysFilemov dl,F.TraceWidthpop bx
 mov ax,Z.ClauseLine; Current - previously shown is in Z.TraceLine
; Produce nothing for equal line number by using zero with full blank suppress.
 mov cx,axxchg cx,Z.TraceLinecmp ax,cxjne @Fz ax@@:call Ax2DsSiCx
 TestReg axjnz @Fmov Z.AxAsNum+(sizeof Zshape.AxAsNum)-1,' '@@:
; Bring small numbers up to length dl by adding leading blanks.
@@:cmp cl,dljae @Finc cxdec simov Bptr[si],' 'jmp @B@@:
; Write from SI onward.
 mov dx,siadd cx,5; Blank-tag-blank.
 mov ah,$DosWriteint 21h; Takes ds:dx:cx to file bx
 mov Z.Tag+2,','; For case of Tag was *-*
 pop cxpop dxpop si
 ret

ShowLine:; Show SS:DX to SS:SI. Advances DX
 mov bx,Z.Stdoutcall ShowClauseId
 pusha
; Alignment may be needed.
 movzx cx,Bptr Z.TraceSourceColumn; Bug havoc potential reduction.
@@:jcxz @Fdec cxcall ShowBlankjmp @B
@@:mov cx,sisub cx,dx
 mov ah,$DosWriteint 21h
 mov dx,si; Ready for next.
 popa
 ret

ShowBlank:pusha
 mov dx,Zshape.Tag; Which is a blank
 mov cx,1
 mov ah,$DosWriteint 21h
 poparet

ClauseSourceEnd:; Records end of a clause in source.
; This is the end of the clause from a source view point - the semicolon,
; linefeed, THEN, ELSE.
 push ScanSipush ax
 sub ScanSi,Zshape.Serialmovzx esi,siadd esi,Z.Relative
 mov eax,esi
; End of previous was beginning of this.
 xchg Z.TraceEnds,esimov Z.TraceSource,esi
; Also need the column.
 sub eax,Z.TraceAlignxchg ax,Z.TraceEndsColumnmov Z.TraceSourceColumn,ax
 pop axpop ScanSiret

;------------------------------------------------------------------------------
; Ax as a number in AxAsNum
; ds and si and cx set up for movsb
;------------------------------------------------------------------------------
Ax2DsSiCx:push axpush bxpush dxmv ds,ss
; Radix always reverts to 10
 mov bx,10xchg bl,Z.Radixz cx
; Convert and store backwards
 mov si,Zshape.AxAsNum + (sizeof Zshape.AxAsNum)
Ax2:z dx; Clear top
 div bx; Divide to get last digit as remainder
 add dl,'0'; Convert to ASCII
; ABCDEF are not immediately after 0123456789
 cmp dl,'9'jbe @Fadd dl,7@@:
 dec simov [si+bp],dlinc cx
 TestReg axjnz Ax2
; Make it even unless going to decimal.
 test cl,1jz @Fcmp bl,10je @Fdec simov Bptr [si+bp],'0'inc cx@@:
 pop dxpop bxpop ax
 ret

